unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, ComCtrls, ExtCtrls, ToolWin, ImgList, StdCtrls, ShellAPI;

{$DEFINE SHISENSHO_TYPEDEF}
  {$I SS_Decl.inc.pas}
{$UNDEF SHISENSHO_TYPEDEF}

type
  TMainForm = class(TForm)
    ToolBar: TToolBar;
    FieldBevel: TBevel;
    ShowMoveTimer: TTimer;
    BNew: TToolButton;
    Symbols: TImageList;
    BHint: TToolButton;
    Spacer1: TToolButton;
    BUndo: TToolButton;
    BRedo: TToolButton;
    Spacer2: TToolButton;
    BPause: TToolButton;
    BScores: TToolButton;
    Spacer3: TToolButton;
    SizeSelect: TComboBox;
    Spacer4: TToolButton;
    InfoPanel: TPanel;
    Clock: TTimer;
    TimeDisp: TLabel;
    InfoMovesLeft: TLabel;
    Bevel1: TBevel;
    Bevel2: TBevel;
    InfoSolvable: TLabel;
    BHelp: TToolButton;
    Spacer5: TToolButton;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FieldMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FieldMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ShowMoveTimerTimer(Sender: TObject);
    procedure BNewClick(Sender: TObject);
    procedure BScoresClick(Sender: TObject);
    procedure SizeSelectChange(Sender: TObject);
    procedure BHintClick(Sender: TObject);
    procedure ClockTimer(Sender: TObject);
    procedure BUndoClick(Sender: TObject);
    procedure BRedoClick(Sender: TObject);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure BPauseClick(Sender: TObject);
    procedure FieldClick(Sender: TObject);
    procedure BHelpClick(Sender: TObject);
  private
    { Private-Deklarationen}
    {$I SS_Decl.inc.pas}
  public
    { Public-Deklarationen}
  end;

var
  MainForm: TMainForm;

implementation

uses Score;

{$R *.DFM}

{$I SS_Logic.inc.pas}
{$I SS_Game.inc.pas}

// ########################################################### MainForm.OnCreate
// Initialisierung
procedure TMainForm.FormCreate(Sender: TObject);
begin
  randomize;
  GraphicsInitialized:=false;
  BufSizeX:=0; BufSizeY:=0;
  SizeSelect.ItemIndex:=3;
  SizeSelectChange(nil);  // InitGraphics, NewGame usw. wird dort erledigt
end;

// ########################################################## MainForm.OnDestroy
// gibt ein paar Bitmaps frei
procedure TMainForm.FormDestroy(Sender: TObject);
begin
  Sprites.Free; Buf.Free; Back.Free;
end;

// ####################################################### ShowMoveTimer.OnTimer
// zeichnet nach abgelaufenem Intervall den Bildschirm neu
procedure TMainForm.ShowMoveTimerTimer(Sender: TObject);
begin
  Redraw;
  ShowMoveTimer.Enabled:=false;
end;

// ################################################################ BNew.OnClick
// startet ein neues Spiel
procedure TMainForm.BNewClick(Sender: TObject);
begin
  NewGame; Redraw; EvaluateGame;
end;

// ############################################################# BScores.OnClick
// Zeigt den Highscore-Dialog an
procedure TMainForm.BScoresClick(Sender: TObject);
begin
  ScoreForm.ShowTileCount:=FieldSizeX*FieldSizeY;
  ScoreForm.ShowModal;
end;

// ######################################################### SizeSelect.OnChange
// ndert die Spielfeldgre
procedure TMainForm.SizeSelectChange(Sender: TObject);
begin
  case SizeSelect.ItemIndex of
    0:begin FieldSizeX:=21; FieldSizeY:=4; end;
    1:begin FieldSizeX:=14; FieldSizeY:=6; end;
    2:begin FieldSizeX:=12; FieldSizeY:=7; end;
    3:begin FieldSizeX:=18; FieldSizeY:=8; end;
    4:begin FieldSizeX:=16; FieldSizeY:=9; end;
    5:begin FieldSizeX:=21; FieldSizeY:=8; end;
  end;
  InitGraphics;
  NewGame; Redraw; EvaluateGame;
end;

// ############################################################### BHint.OnClick
// gibt einen Hinweis bzw. fhrt in aus
procedure TMainForm.BHintClick(Sender: TObject);
begin
  if HintAvailable then begin
    MoveCheck(HintX1,HintY1,HintX2,HintY2,true);
    DoMove(HintX1,HintY1,HintX2,HintY2);
    exit;
  end;
  if MovesLeft=0 then exit;
  Enlightenment(HintX1,HintY1); Enlightenment(HintX2,HintY2);
  Flip; HintAvailable:=true;
  TimeBase:=GetTickCount;TimeStart:=TimeSec+HintPenalty;
  TimeSec:=TimeStart;UpdateTime;
end;

// ############################################################### Clock.OnTimer
// Berechnet die Spielzeit und lt sie ggf. neu anzeigen
procedure TMainForm.ClockTimer(Sender: TObject);
var NewTime:integer;
begin
  NewTime:=TimeStart+(integer(GetTickCount)-TimeBase) DIV 1000;
  if (NewTime<>TimeSec) then begin
    TimeSec:=NewTime;
    UpdateTime;
  end;
end;

// ################################################################## UpdateTime
// Zeichnet das Zeit-Panel neu
procedure TMainForm.UpdateTime;
var min,sec:integer; s:string[5];
begin
  s[0]:=#5; s[3]:=':';
  min:=TimeSec DIV 60;
  s[1]:=char(48+min DIV 10);
  s[2]:=char(48+min MOD 10);
  sec:=TimeSec MOD 60;
  s[4]:=char(48+sec DIV 10);
  s[5]:=char(48+sec MOD 10);
  TimeDisp.Caption:=s;
end;

// ############################################################### BUndo.OnClick
// nimmt einen Spielschritt zurck
procedure TMainForm.BUndoClick(Sender: TObject);
begin
  if HistPos<1 then exit;
  with History[HistPos,1] do FieldData[X,Y]:=Sprite;
  with History[HistPos,2] do FieldData[X,Y]:=Sprite;
  dec(HistPos);
  Redraw; EvaluateGame;
end;

// ############################################################### BRedo.OnClick
// fhrt einen Spielschritt wieder aus
procedure TMainForm.BRedoClick(Sender: TObject);
begin
  if HistPos>=HistMax then exit;
  inc(HistPos);
  with History[HistPos,1] do FieldData[X,Y]:=0;
  with History[HistPos,2] do FieldData[X,Y]:=0;
  Redraw; EvaluateGame;
end;

// ############################################################ MainForm.OnKeyUp
// sorgt fr die Erreichbarkeit der Menbefehle per Strg-Tastenkombi
procedure TMainForm.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key=112 then BHelpClick(nil);
  if not (ssCtrl IN Shift) then exit;
  case Key of
    byte('N'):BNewClick(nil);
    byte('H'):BHintClick(nil);
    byte('Z'):BUndoClick(nil);
    byte('Y'):BRedoClick(nil);
    byte('P'):BPauseClick(nil);
  end;
end;

// ############################################################## BPause.OnClick
// aktiviert den Pausenmodus
procedure TMainForm.BPauseClick(Sender: TObject);
var X,Y:integer;
begin
  if Paused then begin FieldClick(nil); exit; end;
  Paused:=true;
  Clock.Enabled:=false;
  TimeStart:=TimeSec;
  Y:=0;while Y<BufSizeY do begin
    X:=0;while X<BufSizeX do begin
      Buf.Canvas.Draw(X,Y,Back);
    inc(X,BackSizeX);end;
  inc(Y,BackSizeY);end;
  DrawBigText('Pause',clYellow);
  Flip;
end;

// ############################################################### Field.OnClick
// hebt den Pausenmodus auf, falls er aktiviert ist
procedure TMainForm.FieldClick(Sender: TObject);
begin
  if not Paused then exit;
  Paused:=false;
  TimeBase:=GetTickCount;
  Clock.Enabled:=true;
  Redraw;
end;

// ############################################################### BHelp.OnClick
// startet die Hilfe-HTML via ShellExecute
procedure TMainForm.BHelpClick(Sender: TObject);
var res:cardinal;
begin
  res:=ShellExecute(Handle,'open','ShisenSho.htm','','',SW_SHOW);
  if res<32 then
    ShowMessage('Sorry, aber ich kann die Hilfe nicht aufrufen.');
end;

end.
