// ##################################################################### NewGame
// erstellt ein neues Spiel
procedure TMainForm.NewGame;
begin
  SelTileX:=0; SelTileY:=0;
  FillChar(FieldData,sizeof(FieldData),0);
  TilesMax:=FieldSizeX*FieldSizeY;
  TilesLeft:=TilesMax;
  HistPos:=0; HistMax:=0;
  HintAvailable:=false;
  GenerateGame_Clever;
  CountTiles(FieldData);
  TimeStart:=0;TimeBase:=GetTickCount;TimeSec:=0;
  Paused:=false;
  Clock.Enabled:=true;UpdateTime;
end;

// ########################################################### GenerateGame_Dumb
// "dummer" Spielfeldgenerator, der die Steine einfach zufllig
// auf dem Spielfeld verteilt
procedure TMainForm.GenerateGame_Dumb;
var X,Y,Sprite,i:integer;
    TilesAvailable:TTileCount;
begin
  // Phase 1: Anzahl zu erstellender Tiles je Sprite ermitteln
  FillChar(TilesAvailable,sizeof(TilesAvailable),0);
  Sprite:=0;
  for i:=(TilesMax SHR 1) downto 1 do begin
    Sprite:=(Sprite MOD 42)+1;
    inc(TilesAvailable[Sprite],2);
  end;
  // Phase 2: Tiles verteilen
  for Y:=1 to FieldSizeY do
    for X:=1 to FieldSizeX do begin
      repeat Sprite:=random(42)+1;
      until TilesAvailable[Sprite]>0;
      dec(TilesAvailable[Sprite]);
      FieldData[X,Y]:=Sprite;
    end;
end;

// ######################################################### GenerateGame_Clever
// "intelligenter" Spielfeldgenerator, der nur lsbare Spiele produziert
// ALGORITHMUS:
   // Es wird zunchst ein 08/15-Zufallsspiel mittels GenerateGame_Dumb
   // erzeugt. Dann wird ber GetMoves-Hinweise so lange testgespielt,
   // bis es nicht mehr weitergeht [1]. Sind an dieser Stelle bereits keine
   // Tiles mehr brig, so ist das Spielfeld lsbar und die Suche
   // beendet. Andernfalls wird eine Liste der noch brigen Tiles er-
   // stellt [2]; anhand dieser werden diese Tiles dann wild durcheinander
   // ausgetauscht, in der Hoffnung, da dabei eine lsbare Situation
   // entsteht [3]. Das wird berprft, indem wieder testgespielt wird,
   // die ganze Geschichte also von vorne beginnt.
procedure TMainForm.GenerateGame_Clever;
var DestField:TPlayField;  // Ziel-Spielfeld (voll besetzt)
    LTilesLeft:integer;  // Anzahl noch briger Tiles
    LostTiles:array of TPoint;
    LTPos:integer;
    X,Y,Sprite:integer;
    ExchangeCount:integer;
    X1,Y1,S1,X2,Y2,S2:integer;
begin
  GenerateGame_Dumb;
  LTilesLeft:=FieldSizeX*FieldSizeY;
  SetLength(LostTiles,LTilesLeft);
  DestField:=FieldData;
  while (LTilesLeft>0) do begin
    // [1] Testspiel
    CountTiles(DestField);  // vorher Tiles auszhlen!!!
    while (LTilesLeft>0) AND (GetMoves>0) do begin
      FieldData[HintX1,HintY1]:=0;
      FieldData[HintX2,HintY2]:=0;
      dec(LTilesLeft,2);
    end;
    if (LTilesLeft>0) then begin
      // [2] Liste briger Tiles erstellen
      LTPos:=-1;
      for Y:=1 to FieldSizeY do for X:=1 to FieldSizeX do begin
        Sprite:=FieldData[X,Y];
        if Sprite>0 then begin
          inc(LTPos);
          LostTiles[LTPos].X:=X;
          LostTiles[LTPos].Y:=Y;
        end;
      end;
      // [3] Tiles vertauschen
      for ExchangeCount:=(LTilesLeft*1) downto 1 do begin
        LTPos:=random(LTilesLeft); X1:=LostTiles[LTPos].X; Y1:=LostTiles[LTPos].Y;
        LTPos:=random(LTilesLeft); X2:=LostTiles[LTPos].X; Y2:=LostTiles[LTPos].Y;
        S1:=FieldData[X1,Y1]; S2:=FieldData[X2,Y2];
        FieldData[X1,Y1]:=S2; FieldData[X2,Y2]:=S1;
        DestField[X1,Y1]:=S2; DestField[X2,Y2]:=S1;
      end;
    end;
  end;
  FieldData:=DestField;
end;

// ################################################################## CountTiles
// zhlt alle Tiles auf dem Spielfeld aus und fllt somit die Arrays
// TileCount und TilePos aus
procedure TMainForm.CountTiles(var F:TPlayField);
var X,Y,Count,Sprite:integer;
begin
  FillChar(TileCount,sizeof(TileCount),0);
  for X:=1 to FieldSizeX do
    for Y:=1 to FieldSizeY do begin
      Sprite:=F[X,Y];
      Count:=TileCount[Sprite]+1;
      TileCount[Sprite]:=Count;
      TilePos[Sprite,Count].X:=X;
      TilePos[Sprite,Count].Y:=Y;
    end;
end;

// ################################################################### MoveCheck
// berprft, ob ein bestimmter Zug regeltechnisch mglich ist und
// zeichnet ggf. den Zugpfad ins Spielfeld ein
// ALGORITHMUS: (lang)
   // Zunchst werden natrlich ungleiche Tile-Sprites abgelehnt, ganz klar.
   // Dann wird fr jedes Tile beprft, wie viele Felder nach oben, unten,
   // links und rechts unbelegt sind. (Dabei werden ggf. auch die Spielfeld-
   // rnder einbezogen, die ja zwangsweise unbelegt sind.)
   // So entstehen um jedes der beiden Tiles A und B jeweils ein mehr oder
   // weniger langer plusfrmiger Korridor aus freien Feldern. Der Bereich,
   // der von den Korridoren beider Tiles eingeschlossen wird, heit G.
   // Innerhalb von G mu nun in horizontaler und vertikaler Richtung
   // ebenfalls nach freien Korridors gesucht werden. Der gltige Korridor
   // mit der krzesten Entfernung, die der Stein zurcklegen mu, ist dann
   // Lsung des Zuges. Gibt es keinen freien Korridor, so ist der Zug
   // unzulssig.
function TMainForm.MoveCheck(X1,Y1,X2,Y2:integer;DrawIt:boolean):boolean;
type TTileEnv=record X,Y,Xmin,Ymin,Xmax,Ymax:integer;end;
       // plusfrmige Korridors um die Ausgangstiles
type TDirection=(Horizontal,Vertical);
type TCorridor=record Dir:TDirection;Valid:boolean;
                      Via1,Via2:TPoint;Dist:integer;end;
       // Daten der Korridors innerhalb G
       // interessant: In Via1 und Via2 werden die Tile-Koordinaten der
       // ntigen Eckpunkte beim Zug gespeichert
// GetTileEnv: ermittelt die Plus-Korridors um ein Tile herum
procedure GetTileEnv(TX,TY:integer;var T:TTileEnv);
  var X,Y:integer;
  begin
    T.X:=TX; T.Y:=TY;
    X:=TX; while (X>0) AND (FieldData[X-1,TY]=0) do dec(X); T.Xmin:=X;
    Y:=TY; while (Y>0) AND (FieldData[TX,Y-1]=0) do dec(Y); T.Ymin:=Y;
    X:=TX; while (X<=FieldSizeX) AND (FieldData[X+1,TY]=0) do inc(X); T.Xmax:=X;
    Y:=TY; while (Y<=FieldSizeY) AND (FieldData[TX,Y+1]=0) do inc(Y); T.Ymax:=Y;
  end;
var A,B,G:TTileEnv;  // Deklaration der Tile-Umgebungen A,B und G
  // GetHorizontalCorridor: berprft einen bestimmten horizontalen Korridor
  procedure GetHorizontalCorridor(Y:integer;var C:TCorridor);
  var X1,X2,X:integer; FTotal,FFree:integer;
  begin
    if A.X<=B.X then begin X1:=A.X;X2:=B.X; end else begin X1:=B.X;X2:=A.X; end;
    FTotal:=0; FFree:=0;
    for X:=X1 to X2 do begin
      if ((X<>A.X) OR (Y<>A.Y)) AND ((X<>B.X) OR (Y<>B.Y)) then inc(FTotal);
      if FieldData[X,Y]=0 then inc(FFree);
    end;
    with C do begin
      Dir:=Horizontal;
      Via1.X:=A.X; Via1.Y:=Y;
      Via2.X:=B.X; Via2.Y:=Y;
      Valid:=(FTotal=FFree);
      Dist:=Abs(A.Y-Y)+Abs(B.Y-Y)+X2-X1+1;
    end;
  end;
  // GetVerticalCorridor: berprft einen bestimmten vertikalen Korridor
  procedure GetVerticalCorridor(X:integer;var C:TCorridor);
  var Y1,Y2,Y:integer; FTotal,FFree:integer;
  begin
    if A.Y<=B.Y then begin Y1:=A.Y;Y2:=B.Y; end else begin Y1:=B.Y;Y2:=A.Y; end;
    FTotal:=0; FFree:=0;
    for Y:=Y1 to Y2 do begin
      if ((X<>A.X) OR (Y<>A.Y)) AND ((X<>B.X) OR (Y<>B.Y)) then inc(FTotal);
      if FieldData[X,Y]=0 then inc(FFree);
    end;
    with C do begin
      Dir:=Horizontal;
      Via1.X:=X; Via1.Y:=A.Y;
      Via2.X:=X; Via2.Y:=B.Y;
      Valid:=(FTotal=FFree);
      Dist:=Abs(A.X-X)+Abs(B.X-X)+Y2-Y1+1;
    end;
  end;
  // DrawCorridor: Zeichnet einen Korridor
  function DrawCorridor(const C:TCorridor):boolean;
  begin
    DrawCorridor:=false;
    if not C.Valid then exit;
    with Buf.Canvas do begin
      MoveTo(A.X*TileSizeX+FieldBorder-TileSizeX SHR 1,
             A.Y*TileSizeY+FieldBorder-TileSizeY SHR 1);
      LineTo(C.Via1.X*TileSizeX+FieldBorder-TileSizeX SHR 1,
             C.Via1.Y*TileSizeY+FieldBorder-TileSizeY SHR 1);
      LineTo(C.Via2.X*TileSizeX+FieldBorder-TileSizeX SHR 1,
             C.Via2.Y*TileSizeY+FieldBorder-TileSizeY SHR 1);
      LineTo(B.X*TileSizeX+FieldBorder-TileSizeX SHR 1,
             B.Y*TileSizeY+FieldBorder-TileSizeY SHR 1);
    end;
    DrawCorridor:=true;
  end;
// Hauptroutine
var X,Y:integer; CurCorridor,BestCorridor:TCorridor;
begin
  // ungleiche Tile-Sprites und doppeltes Null-Tile abweisen
  if (FieldData[X1,Y1]<>FieldData[X2,Y2]) OR (FieldData[X1,Y1]+FieldData[X2,Y2]<1) then begin
    MoveCheck:=false;
    exit;
  end;
  // A- und B-Korridors ermitteln
  GetTileEnv(X1,Y1,A);
  GetTileEnv(X2,Y2,B);
  // Ausmae des G-Bereiches ermitteln
  if A.Xmin>B.Xmin then G.Xmin:=A.Xmin else G.Xmin:=B.Xmin;
  if A.Ymin>B.Ymin then G.Ymin:=A.Ymin else G.Ymin:=B.Ymin;
  if A.Xmax<B.Xmax then G.Xmax:=A.Xmax else G.Xmax:=B.Xmax;
  if A.Ymax<B.Ymax then G.Ymax:=A.Ymax else G.Ymax:=B.Ymax;
  // BestCorridor: noch gar keiner
  BestCorridor.Valid:=false;
  BestCorridor.Dist:=maxint;
  // horizontale Korridors suchen
  for Y:=G.Ymin to G.Ymax do begin
    GetHorizontalCorridor(Y,CurCorridor);
    if CurCorridor.Valid AND (CurCorridor.Dist<BestCorridor.Dist) then
      BestCorridor:=CurCorridor;
  end;
  // vertikale Korridors suchen
  for X:=G.Xmin to G.Xmax do begin
    GetVerticalCorridor(X,CurCorridor);
    if CurCorridor.Valid AND (CurCorridor.Dist<BestCorridor.Dist) then
      BestCorridor:=CurCorridor;
  end;
  // Endauswertung
  if DrawIt AND DrawCorridor(BestCorridor) then Flip;
  MoveCheck:=BestCorridor.Valid;
end;

// #################################################################### GetMoves
// ermittelt die Anzahl gltiger Zge, indem alle Tiles mit gleichen
// Sprites mittels MoveCheck auf Verbindbarkeit berprft werden;
// als Nebenprodukt entsteht dabei ein Zugvorschlag
function TMainForm.GetMoves:integer;
var Moves:integer;
    Sprite,Max,a,b:integer;
begin
  Moves:=0;
  for Sprite:=1 to 42 do begin
    max:=TileCount[Sprite];
    for a:=1 to (max-1) do
      for b:=(a+1) to max do
        if MoveCheck(TilePos[Sprite,a].X,TilePos[Sprite,a].Y,
                     TilePos[Sprite,b].X,TilePos[Sprite,b].Y,false)
        then begin
          inc(Moves);
          HintX1:=TilePos[Sprite,a].X; HintY1:=TilePos[Sprite,a].Y;
          HintX2:=TilePos[Sprite,b].X; HintY2:=TilePos[Sprite,b].Y;
        end;
  end;
  GetMoves:=Moves;
end;

// #################################################################### Solvable
// berprft ein Spielfeld auf Lsbarkeit, indem solange GetMoves-
// Zugvorschlge ausgefhrt werden, bis es nicht mehr weitergeht;
// wenn an diesem Punkt das Spielfeld ohnehin leer ist, dann ist die
// Spielsituation garantiert lsbar
function TMainForm.Solvable:boolean;
var OriginalField:TPlayField;
    LTilesLeft:integer;
begin
  LTilesLeft:=TilesLeft;
  if LTilesLeft=0 then begin Solvable:=true;exit;end;
  OriginalField:=FieldData;
  while (LTilesLeft>0) AND (GetMoves>0) do begin
    FieldData[HintX1,HintY1]:=0;
    FieldData[HintX2,HintY2]:=0;
    dec(LTilesLeft,2);
  end;
  FieldData:=OriginalField;
  Solvable:=(LTilesLeft=0);
end;

