unit Score;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Buttons;

const ScoreFile='ShisenSho.hi';  // Spielstand-Datei

type TScoreboardEntry=record  // ########## ein Hall-of-Fame-Eintrag ###########
                        Name:string[31];  // Spielername
                        SizeX,SizeY:integer;  // konkrete Spielfeldgre X/Y
                        Time:integer;  // Spielzeit in Sekunden
                        Date:integer;  // Datum
                      end;
     TScoreboardData=record  // ### eine Hall of Fame (fr 1 Spielfeldgre) ###
                       TileNum:integer;  // Anzahl der Steine = Spielfeldgre
                       Data:array[1..11]of TScoreboardEntry;  // Eintrge
                     end;

type TScoreboardControls=record // ein Satz aus HoF-eintragsanzeigenden Controls
                           Nr,Name,Size,Time,Date:TLabel;
                         end;
// zur Erklrung: Laut Formularansicht besteht das Formular zu 99% aus Luft.
//                Das ist auch so gewollt, denn die 10 TLabels, welche dann die
//                Hall-of-Fame-Eintrge anzeigen, werden erst zur Laufzeit
//                erstellt, damit man sie schn praktisch in einem Array
//                verwalten kann.

type
  TScoreForm = class(TForm)
    CRank: TLabel;
    CName: TLabel;
    CSize: TLabel;
    CTime: TLabel;
    CDate: TLabel;
    Bevel1: TBevel;
    BOK: TBitBtn;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    { Private-Deklarationen}
    FormBackground:TBitmap;  // Hintergrund-Bitmap
    Marked:integer;  // welcher HoF-Eintrag ist als "frisch" markiert?
    Scores:array of TScoreboardData;  // HoF-Datenstze
    SCtl:array[1..10]of TScoreboardControls;  // Anzeige-Controls
    // Beschreibungen der Methoden siehe implementation-Teil
    function GetScoreIndex(TileCount:integer):integer;
    procedure SaveHighscores;
    procedure LoadHighscores;
  public
    { Public-Deklarationen}
    ShowTileCount:integer;  // Steinzahl, fr die die Hall of Fame angezeigt werden soll
    function AddScore(const PlayerName:string;SX,SY,_Time:integer):boolean;
  end;

var
  ScoreForm: TScoreForm;

implementation

const hex:string[16]='0123456789ABCDEF';

{$R *.DFM}
{$R COOLCTLS.RES}  // fr die Hintergrund-Bitmap

// ########################################################## ScoreForm.OnCreate
// initialisiert die HoF-Controls und den Datenbereich
procedure TScoreForm.FormCreate(Sender: TObject);
var i,y:integer;
begin
  y:=28;
  for i:=1 to 10 do begin
    with SCtl[i] do begin
      Nr:=TLabel.Create(Self);
      with Nr do begin
        Parent:=Self;
        Top:=y; Left:=8;
        AutoSize:=false; Width:=17;
        ShowAccelChar:=false; Transparent:=true;
        Alignment:=taRightJustify;
        Caption:=IntToStr(i)+'.';
      end;
      Name:=TLabel.Create(Self);
        with Name do begin Parent:=Self; Top:=y; Left:=32;
        ShowAccelChar:=false; Transparent:=true;
        Caption:='Shisen-Sho'; end;
      Size:=TLabel.Create(Self);
        with Size do begin Parent:=Self; Top:=y; Left:=136;
        ShowAccelChar:=false; Transparent:=true;
        Caption:='00x0'; end;
      Time:=TLabel.Create(Self);
        with Time do begin Parent:=Self; Top:=y; Left:=184;
        ShowAccelChar:=false; Transparent:=true;
        Caption:='00:00'; end;
      Date:=TLabel.Create(Self);
        with Date do begin Parent:=Self; Top:=y; Left:=224;
        ShowAccelChar:=false; Transparent:=true;
        Caption:='00.00.0000'; end;
      inc(y,18);
    end;
  end;
  Marked:=0;
  with Canvas do begin
    Pen.Style:=psClear;
    Brush.Style:=bsSolid;
  end;
  FormBackground:=TBitmap.Create;
  FormBackground.LoadFromResourceName(MainInstance,'COOLBACK');
  SetLength(Scores,0);
  ShowTileCount:=84;
  LoadHighscores;
end;

// ######################################################### ScoreForm.OnDestroy
// gibt die HoF-Labels wieder frei
procedure TScoreForm.FormDestroy(Sender: TObject);
var i:integer;
begin
  SaveHighscores;
  for i:=1 to 10 do
    with SCtl[i] do begin
      Nr.Free;
      Name.Free;
      Size.Free;
      Time.Free;
      Date.Free;
    end;
  FormBackground.Free;
end;

// ############################################################ ScoreForm.OnHide
// Sorgt dafr, da beim nxten Mal kein Eintrag hervorgehoben wird
procedure TScoreForm.FormHide(Sender: TObject);
begin
  Marked:=0;
end;

// ########################################################### ScoreForm.OnPaint
// Zeichnet die Hintergrund-Bitmap und hebt evtl. einen Eintrag hervor
procedure TScoreForm.FormPaint(Sender: TObject);
var x,y:integer;
begin
  y:=0;
  while y<248 do begin
    x:=0;
    while x<288 do begin
      Canvas.Draw(x,y,FormBackground);
      inc(x,64);
    end;
    inc(y,64);
  end;
  if Marked<1 then exit;
  with Canvas do begin
    Brush.Color:=$EDD0D0+(10-Marked)*$020000;
    FillRect(Rect(6,18*Marked+8,282,18*Marked+25));
  end;
end;

// ############################################################ ScoreForm.OnShow
// Trgt die entsprechende Highscore-Liste in die VCL-Controls ein
procedure TScoreForm.FormShow(Sender: TObject);
type str2=string[2];
var i,t:integer;
    year,month,day:word;
function st2(i:integer):str2;begin st2:=char(48+i DIV 10)+char(48+i MOD 10);end;
begin
  with Scores[GetScoreIndex(ShowTileCount)] do
    for i:=1 to 10 do
      with SCtl[i] do
        if Data[i].Date=0 then begin
          Name.Caption:='';
          Size.Caption:='';
          Time.Caption:='';
          Date.Caption:='';
        end else begin
          Name.Caption:=Data[i].Name;
          Size.Caption:=IntToStr(Data[i].SizeX)+'x'+IntToStr(Data[i].SizeY);
          t:=Data[i].Time;
          Time.Caption:=st2(t DIV 60)+':'+st2(t MOD 60);
          DecodeDate(Data[i].Date+0.5,year,month,day);
          Date.Caption:=st2(day)+'.'+st2(month)+'.'+st2(year DIV 100)+st2(year MOD 100);
        end;
end;

// #################################################################### AddScore
// wird vom Host aufgerufen, um eine neue Punktzahl einzutragen; dabei
// wird auch alles fr einen folgenden ShowModal-Aufruf vorbereitet
function TScoreForm.AddScore(const PlayerName:string;SX,SY,_Time:integer):boolean;
var temp:TScoreboardEntry;
begin
  ShowTileCount:=SX*SY;
  with Scores[GetScoreIndex(ShowTileCount)] do begin
    Marked:=11;
    with Data[11] do begin
      Name:=PlayerName;
      SizeX:=SX; SizeY:=SY;
      Time:=_Time;
      Date:=trunc(Now);
    end;
    while (Marked>1) AND (Data[Marked-1].Time>Data[Marked].Time) do begin
      dec(Marked);
      temp:=Data[Marked];
      Data[Marked]:=Data[Marked+1];
      Data[Marked+1]:=temp;
    end;
  end;
  if Marked>10 then Marked:=0;
  AddScore:=(Marked<=10);
end;

// ############################################################## SaveHighscores
// Speichert die Highscores
procedure TScoreForm.SaveHighscores;
var t:system.text; s:string[127];
    Index,SubIndex:integer; b:byte;
procedure AddByte(b:byte);begin s:=s+hex[1+(b SHR 4)]+hex[1+(b AND 15)]; end;
procedure AddWord(w:word);begin AddByte(w SHR 8);AddByte(w AND 255); end;
procedure AddLong(l:cardinal);begin AddWord(l SHR 16);AddWord(l AND $FFFF); end;
begin
  AssignFile(t,ScoreFile); Rewrite(t);
  for Index:=0 to High(Scores) do
    with Scores[Index] do begin
      writeln(t,'@',TileNum);
      for SubIndex:=1 to 10 do
        with Data[SubIndex] do
          if Date>0 then begin
            s:='#';
            for b:=0 to length(Name) do
              AddByte(byte(Name[b]));
            AddByte(SizeX);
            AddByte(SizeY);
            AddWord(Time);
            AddLong(Date);
            writeln(t,s);
          end;
    end;
  CloseFile(t);
end;

// ############################################################## LoadHighscores
// ldt die Highscores
procedure TScoreForm.LoadHighscores;
var t:system.text; s:string[127]; c:^char;
    Index,SubIndex:integer; m,b:byte;
function GetByte:byte;var a,b:char;begin a:=upcase(c^);inc(c); b:=upcase(c^);inc(c);
  GetByte:=(pos(a,hex)-1) SHL 4 OR (pos(b,hex)-1);end;
function GetWord:word;begin GetWord:=GetByte SHL 8 OR GetByte;end;
function GetLong:cardinal;begin GetLong:=GetWord SHL 16 OR GetWord;end;
begin
  AssignFile(t,ScoreFile);
  {$I-} Reset(t); {$I+}
  if IOresult<>0 then exit;
  Index:=High(Scores);SubIndex:=1;
  while not EOF(t) do begin
    FillChar(s,sizeof(s),0);
    readln(t,s);
    if s[1]='@' then begin
      Index:=GetScoreIndex(StrToInt(copy(s,2,3)));
      SubIndex:=1;
    end else if (s[1]='#') AND (Index>=0) AND (SubIndex<11) then
      with Scores[Index] do begin
        c:=@s[2];
        with Data[SubIndex] do begin
          m:=GetByte;
          if (m>0) AND (m<sizeof(Name)) then begin
            Name[0]:=char(m);
            for b:=1 to m do
              Name[b]:=char(GetByte);
            SizeX:=GetByte;
            SizeY:=GetByte;
            Time:=GetWord;
            Date:=GetLong;
          end;
        end;
        inc(SubIndex);
      end;
  end;
  CloseFile(t);
end;

// ############################################################### GetScoreIndex
// Sucht nach der Highscore-Liste mit einer bestimmten Spielfeldgre
// (=Steinanzahl) und gibt dessen Nummer zurck. Existiert die
// gewnschte Gre noch nicht, so wird ein neuer Eintrag angelegt.
function TScoreForm.GetScoreIndex(TileCount:integer):integer;
var Count,Found:integer;
begin
  Found:=-1;
  for Count:=0 to (High(Scores)) do
    if Scores[Count].TileNum=TileCount then
      Found:=Count;
  if Found>=0 then begin GetScoreIndex:=Found;exit;end;
  Found:=High(Scores)+1;
  SetLength(Scores,Found+1);
  with Scores[Found] do begin
    for Count:=1 to 11 do
      with Data[Count] do begin
        Name:='Shisen-Sho';
        SizeX:=0; SizeY:=0;
        Time:=maxint; Date:=0;
      end;
    TileNum:=TileCount;
  end;
  GetScoreIndex:=Found;
end;

end.
